'''
Based on https://github.com/tensorflow/tensorflow/blob/master/tensorflow/examples/tutorials/layers/cnn_mnist.py:

A Convolutional Network implementation example using TensorFlow library.
This example is using the MNIST database of handwritten digits
(http://yann.lecun.com/exdb/mnist/)

Author: TensorFlow
Project: https://github.com/tensorflow/tensorflow

Modifications:

*****************************************************************

Licensed Materials - Property of IBM

(C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.

US Government Users Restricted Rights - Use, duplication or
disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

*****************************************************************
'''

import sys
import os
import shutil
import argparse

import tensorflow as tf
if len(sys.argv) > 1 and sys.argv[1] == '--eager':
    tf.enable_eager_execution()
import ddl


import dataset

tf.logging.set_verbosity(tf.logging.INFO)


def cnn_model_fn(features, labels, mode):
    # Input Layer
    # Reshape X to 4-D tensor: [batch_size, width, height, channels]
    # MNIST images are 28x28 pixels, and have one color channel
    input_layer = tf.reshape(features, [-1, 28, 28, 1])

    # Convolutional Layer #1
    # Computes 32 features using a 5x5 filter with ReLU activation.
    # Padding is added to preserve width and height.
    # Input Tensor Shape: [batch_size, 28, 28, 1]
    # Output Tensor Shape: [batch_size, 28, 28, 32]
    conv1 = tf.layers.conv2d(
        inputs=input_layer,
        filters=32,
        kernel_size=[5, 5],
        padding="same",
        activation=tf.nn.relu)

    # Pooling Layer #1
    # First max pooling layer with a 2x2 filter and stride of 2
    # Input Tensor Shape: [batch_size, 28, 28, 32]
    # Output Tensor Shape: [batch_size, 14, 14, 32]
    pool1 = tf.layers.max_pooling2d(inputs=conv1, pool_size=[2, 2], strides=2)

    # Convolutional Layer #2
    # Computes 64 features using a 5x5 filter.
    # Padding is added to preserve width and height.
    # Input Tensor Shape: [batch_size, 14, 14, 32]
    # Output Tensor Shape: [batch_size, 14, 14, 64]
    conv2 = tf.layers.conv2d(
        inputs=pool1,
        filters=64,
        kernel_size=[5, 5],
        padding="same",
        activation=tf.nn.relu)

    # Pooling Layer #2
    # Second max pooling layer with a 2x2 filter and stride of 2
    # Input Tensor Shape: [batch_size, 14, 14, 64]
    # Output Tensor Shape: [batch_size, 7, 7, 64]
    pool2 = tf.layers.max_pooling2d(inputs=conv2, pool_size=[2, 2], strides=2)

    # Flatten tensor into a batch of vectors
    # Input Tensor Shape: [batch_size, 7, 7, 64]
    # Output Tensor Shape: [batch_size, 7 * 7 * 64]
    pool2_flat = tf.reshape(pool2, [-1, 7 * 7 * 64])

    # Dense Layer
    # Densely connected layer with 1024 neurons
    # Input Tensor Shape: [batch_size, 7 * 7 * 64]
    # Output Tensor Shape: [batch_size, 1024]
    dense = tf.layers.dense(inputs=pool2_flat, units=1024, activation=tf.nn.relu)

    # Add dropout operation; 0.6 probability that element will be kept
    dropout = tf.layers.dropout(
        inputs=dense, rate=0.4, training=mode == tf.estimator.ModeKeys.TRAIN)

    # Logits layer
    # Input Tensor Shape: [batch_size, 1024]
    # Output Tensor Shape: [batch_size, 10]
    logits = tf.layers.dense(inputs=dropout, units=10)

    predictions = {
        # Generate predictions (for PREDICT and EVAL mode)
        "classes": tf.argmax(input=logits, axis=1),
        # Add `softmax_tensor` to the graph. It is used for PREDICT and by the
        # `logging hook`
        "probabilities": tf.nn.softmax(logits, name="softmax_tensor"),
    }

    labels = tf.argmax(labels, axis=1)

    # Calculate Loss (for both TRAIN and EVAL modes)
    loss = tf.losses.sparse_softmax_cross_entropy(labels=labels, logits=logits)

    # Calculate accuracy for EVAL mode
    accuracy = tf.metrics.accuracy( labels=labels,
                                    predictions=predictions["classes"],
                                    name="accuracy")

    # Logging hook to see progress as model trains
    tensors_to_log = {"accuracy": accuracy[1],
                      "loss"    : loss}
    logging_hook  = tf.train.LoggingTensorHook(tensors=tensors_to_log,
                                               every_n_iter=500)

    # PREDICT mode
    if mode == tf.estimator.ModeKeys.PREDICT:
        return tf.estimator.EstimatorSpec(mode=mode, predictions=predictions)

    # Configure the Training Op (for TRAIN mode)
    if mode == tf.estimator.ModeKeys.TRAIN:
        optimizer = tf.train.GradientDescentOptimizer(learning_rate=0.001)
        train_op = optimizer.minimize(  loss=loss,
                    global_step=tf.train.get_global_step())
        if ddl.rank() == 0:
            return tf.estimator.EstimatorSpec(mode=mode, loss=loss, train_op=train_op,
                    training_hooks=[logging_hook])
        else:
            return tf.estimator.EstimatorSpec(mode=mode, loss=loss, train_op=train_op)

    # EVAL mode
    return tf.estimator.EstimatorSpec(
        mode=mode, loss=loss, eval_metric_ops={"accuracy" : accuracy})

def main ():
    data_dir  = "/tmp/mnist_convnet_model_data"+str(ddl.rank())
    model_dir = "/tmp/mnist_convnet_model"+str(ddl.rank())
    #Clear contents of model_dir prior to going forward to make sure
    #training starts fresh each run
    if os.path.isdir(model_dir):
        shutil.rmtree(model_dir)

    #Define a run config which save chkpts more frequently
    batch_size = 100
    run_config = tf.estimator.RunConfig(
                        save_checkpoints_steps=100,
                        save_checkpoints_secs=None)

    #Create the estimator
    mnist_classifier = tf.estimator.Estimator(
                        model_fn=cnn_model_fn,
                        model_dir=model_dir,
                        config=run_config)

    # Train function
    def train_input_fn ():
        ds, _ = dataset.train(data_dir, (28, 28, 1))
        ds = ds.shard(ddl.size(), ddl.rank())
        ds = ds.cache().shuffle(buffer_size=50000).batch(batch_size).repeat()
        return ds

    train_spec = tf.estimator.TrainSpec(input_fn=train_input_fn,
                                        max_steps=FLAGS.num_iterations,
                                        hooks={ddl.DDLGlobalVariableHook()})

    # Evaluate function
    def eval_input_fn ():
        ds, _ = dataset.test(data_dir, (28, 28, 1))
        ds = ds.shard(ddl.size(), ddl.rank())
        ds = ds.batch(batch_size)
        return ds

    eval_spec  = tf.estimator.EvalSpec(input_fn=eval_input_fn)

    #Train and Evaluate
    eval_results = tf.estimator.train_and_evaluate(mnist_classifier,
                                                   train_spec, eval_spec)

    print ("Eval Results " + str(ddl.rank()) + str(" ") + str(eval_results))


    if ddl.rank() == 0:
        def test_input_fn ():
            ds, _ = dataset.test(data_dir, (28, 28, 1))
            ds = ds.batch(batch_size)
            return ds

        full_result = mnist_classifier.evaluate(input_fn=test_input_fn)
        print ("Test accuracy: " + str(full_result["accuracy"]))

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--num_iterations', type=int,
                        default=20000,
                        help='Max number of training iterations')
    FLAGS, unparsed = parser.parse_known_args()
    main()
